;;; ****************************************************************

;;; Beginning of figures for Appendix F

;;; Figure F.1 : page 493

(define stream-car
  (lambda (s)
    (car (s))))

(define stream-cdr
  (lambda (s)
    (cdr (s))))

(define make-stream
  (lambda (car-val th)
    (lambda ()
      (cons car-val (let ((cdr-val #f))
                      (lambda ()
                        (if (pair? cdr-val)
                            cdr-val
                            (begin (set! cdr-val ((th))) cdr-val))))))))

(define the-null-stream
  (make-stream 'end-of-stream (lambda () the-null-stream)))

(define stream-null?
  (lambda (s)
    (eq? (stream-car s) 'end-of-stream)))

;;; Figure F.2 : page 494

(define stream-for-each
  (lambda (proc stream)
    (letrec
      ((loop
         (lambda (stream)
           (if (not (stream-null? stream))
               (begin
                 (proc (stream-car stream))
                 (loop (stream-cdr stream)))))))
      (loop stream))))

(define make-input-stream
  (lambda ()
    (let ((char (read-char)))
      (if (eof-object? char)
          the-null-stream
          (make-stream char make-input-stream)))))

;;; Figure F.3 : page 494

(define make-char-seq make-stream)

(define char-seq-head stream-car)

(define char-seq-tail stream-cdr)

(define make-token-seq make-stream)

(define token-seq-head stream-car)

(define token-seq-tail stream-cdr)

;;; Figure F.4 : page 495 

(define eval-print
  (lambda (tree)
    (write tree)))

(define eval-print
  (lambda (tree)
    (let ((result (eval-exp tree init-env)))
      (if (not (or (define? tree)
                   (definearray? tree)
                   (varassign? tree)
                   (arrayassign? tree)))
          (write result)))))

(define read-eval-print
  (lambda ()
    (display "--> ")
    (force-output)
    (stream-for-each
      (lambda (tree)
        (eval-print tree)
        (newline)
        (display "--> ")
	(force-output))
      (parse-token-seq parse-semicolon-terminated-form
        (scan-char-seq scanner-start-state (make-input-stream))))))

;;; Figure F.5 : page 496

(define parse-semicolon-terminated-form
  (lambda (token)
    (process-nt parse-form
      (goto-parser-state parse-terminated-by))))

(define parse-terminated-by
  (lambda (token)
    (case (token->class token)
      ((semicolon)
       (check/drop 'semicolon (reduce 'chain)))
      ((end)
       (reduce 'chain))
      (else (error "Invalid terminator:" token)))))

(define parse-token-seq
  (lambda (start-state token-seq)
    (let ((token (token-seq-head token-seq)))
      (if (or (eq? (token->class token) 'end)
              (eq? (token->class token) 'end-marker))
          the-null-stream
          (variant-case (parse-once start-state token-seq)
            (parser-answer (tree unparsed)
              (make-stream tree
                (lambda ()
                  (parse-token-seq start-state unparsed)))))))))

;;; End of figures for Appendix F
