;; $Id: array.scm,v 1.11 1999/04/15 18:17:15 leavens Exp $

;;; Figure 6.1.1 : page 181 %%%%% New array ADT for new direct arrays %%%%%%

(define make-array
  ; TYPE: (-> (number) (array T)) ; unsound!
  (lambda (length)
    (let ((array (make-vector (+ length 1))))
      (vector-set! array 0 '*array*)
      array)))

(define array?
  ; TYPE: (-> (datum) boolean)
  (lambda (x)
    (and (vector? x)
	 (> (vector-length x) 0)
	 (eq? (vector-ref x 0) '*array*))))

(define array-ref
  ; TYPE: (-> ((array T) number) T)
  (lambda (array index)
    (if (array? array)
	(if (or (< index 0) (>= (+ index 1) (vector-length array)))
	    (error "array-ref: Index out of range:" index)
	    (vector-ref array (+ index 1)))
	(error "array-ref: attempt to index into non-array: " array))))

(define array-set!
  ; TYPE: (-> ((array T) number T) void)
  (lambda (array index value)
    (if (array? array)
	(if (or (< index 0) (>= (+ index 1) (vector-length array)))
	    (error "array-set!: Index out of range:" index)
	    (vector-set! array (+ index 1) value))
	(error "array-set!: attempt to set non-array: " array))))

(define array-whole-set!
  ; TYPE: (-> ((array T) (array T)) void)
  (lambda (dest-array source-array)
    (if (and (array? dest-array) (array? source-array))
	(let ((source-len (vector-length source-array)))
	  (if (> source-len (vector-length dest-array))
	      (error "Array too long for assignment: " source-array)
	      (letrec 
		  ((loop (lambda (n)
			   (if (< n source-len)
			       (begin
				 (vector-set! dest-array
					      n
					      (vector-ref source-array n))
				 (loop (+ n 1)))))))
		(loop 1))))
	(error "array-whole-set!: both arguments are not arrays: "
	       dest-array source-array))))

(define array-copy
  ; TYPE: (-> ((array T)) (array T))
  (lambda (array)
    (if (array? array)
	(let ((new-array (make-array (- (vector-length array) 1))))
	  (array-whole-set! new-array array)
	  new-array)
	(error "array-copy: attempt to copy non-array: " array))))

;; the following is added locally

(define array-length
  ; TYPE: (-> ((array T)) number)
  (lambda (array)
    (if (array? array)
	(- (vector-length array) 1)
	(error "array-length: application to non-array: " array))))
