;; $Id: ch6-5-2.scm,v 1.12 1998/11/12 22:53:20 leavens Exp $

;; Code from the EOPL chapter 6 revision, Figure 6.5.2, as further revised.
;; (call-by-name with arrays in the indirect model)

;;; Grammar:
;;          see the file ch6-base.scm for details,
;;          except that the production for <operand> is replaced by
;;                <operand> ::= <varref>
;;                            | <array-exp> [<exp>]
;;                            | <exp>

(load-from-lib "ch6-base.scm")

;;; Domains for this interpreter:
;;;
;;; Environment = <var> -> Denoted-Value
;;; Denoted-Value = L-Value + Thunk
;;; Expressed-Value = Number + Procedure + Array
;;;                   + List(Expressed-Value) + Void
;;; Procedure = prim-proc + closure
;;; Array(T) = {Cell(T)}*
;;; L-Value = Cell(Expressed-Value) + Array-Element(Expressed-Value)
;;; Thunk = () -> L-Value

(load-quietly-from-lib "ch6-5-2-denoted-value.scm")
(load-quietly-from-lib "ch6-5-2-l-value.scm")
(load-quietly-from-lib "ch6-1-3-expressed-value.scm")


;;; *Modified* from Figure 6.5.2, page 204 and Figure 6.1.3, page 183

(define denoted->expressed
  ;; TYPE: (-> (Denoted-Value) Expressed-Value)
  (lambda (den-val)
    (let ((l-val (denoted->l-value den-val)))
      (cond
        ((cell? l-val) (cell-ref l-val))
        ((ae? l-val)
         (array-ref (ae->array l-val) (ae->index l-val)))
        (else (error "Can't dereference denoted value: " l-val))))))

(define expressed->denoted
  ;; TYPE: (-> (Expressed-Value) Denoted-Value)
  (lambda (ev)
    (l-value->denoted (make-cell ev))))

(define denoted-value-assign!
  ;; TYPE: (-> (Denoted-Value Expressed-Value) void)
  (lambda (den-val exp-val)
    (let ((l-val (denoted->l-value den-val)))
      (cond
        ((cell? l-val) (cell-set! l-val exp-val))
        ((ae? l-val)
	 (array-element-assign!  ; changed to better support direct model!
	  (ae->array l-val)
	  (ae->index l-val)
	  exp-val))
        (else (error "Can't assign to denoted value: " l-val))))))

(define do-letarray
  ;; TYPE: (-> (Expressed-Value) Denoted-Value)
  (lambda (n)
    (if (not (<= 0 n))  ; added error check vs. text
	(error "do-letarray: bad index")
	(l-value->denoted (make-cell (array->expressed (make-array n)))))))

(define eval-array-exp
  ;; TYPE: (-> (parsed-exp Environment) Array)
  (lambda (exp env)
    (eval-exp exp env)))

;; addition to figure 6.1.3

(define array-element-assign!
  ;; TYPE: (-> ((Array Expressed-Value) number Expressed-Value) void)
  array-set!)

;;; *Modified* from Figure 6.5.1 : page 203 

(define eval-rand   ;; eval-rand in 6.5.2 is something else, this fits as usual
  ; TYPE: (-> (parsed-exp Environment) Denoted-Value)
  (lambda (rand env)
    (variant-case rand
      (varref (var)
	 (let ((den-val (apply-env env var)))
	   (if (thunk? den-val)
	       den-val
	       (make-thunk rand env))))
      (else (make-thunk rand env)))))
