;; $Id: ch9-2-3.scm,v 1.1 1999/08/11 16:21:33 leavens Exp $

;; Code for EOPL section 9.2 using the chararacter string parser.

(load-from-lib "ch9-base.scm")

;;; Top-level Forms

(load-from-lib "eval-form-cont.scm")

;;; Grammar:
;;	    <exp> ::= <integer-literal>
;;		    | <varref>
;;		    | <operator> <operands>
;;	            | proc <varlist> <exp>
;;                  | if <exp> then <exp> else <exp>
;;                  | letcont <var> in <exp>
;;	<operator> ::= <varref> | (<exp>)
;;	<operands> ::= ( )
;;	            | (<operand> {, <operand>}*)
;;	 <operand> ::= <exp>
;;	  <varref> ::= <var>
;;	<decls> ::= <decl> {;<decl>}*
;;	 <decl> ::= <var> = <exp>
;;	<varlist> ::= ( ) | ( <vars> )
;;	 <vars> ::= <var> {,<var>}*

;;; Evaluation of expressions

;;; Domains for this interpreter:
;;;
;;; Environment = <var> -> Denoted-Value
;;; Denoted-Value = Number + Procedure
;;; Expressed-Value = Number + Procedure
;;;                   + Pair(Expressed-Value Expressed-Value) + Null
;;; Procedure = prim-proc + closure + abort-proc + continuation

;;; Anonymous figure : page 307, 308

(define-record continuation (cont))
(define-record letcont (var body))
(define-record abort (exp))  ;; recognized by the parser as reserved syntax!

;;; Figure 9.2.3 : page 301

(define eval-exp
  (lambda (exp env k)
    (variant-case exp
      (lit (datum) 
        (apply-continuation k datum))
      (varref (var) 
        (apply-continuation k (apply-env env var)))
      (app (rator rands) 
        (eval-exp rator env 
          (make-proc-valcont rands env k)))
      (if (test-exp then-exp else-exp)
        (eval-exp test-exp env
          (make-test-valcont then-exp else-exp env k)))
      (proc (formals body)
        (apply-continuation k 
          (make-closure formals body env)))
      (letcont (var body)
        (eval-exp body
          (extend-env (list var) (list (make-continuation k)) env)
          k))
      (else (error "Invalid abstract syntax:" exp)))))

(define eval-rands
  (lambda (rands env k)
    (if (null? rands)
        (apply-continuation k '())
        (eval-exp (car rands) env
          (make-first-valcont rands env k)))))

;; for apply-proc, see below

;;; Figure 9.2.5 : page 304

(define-record final-valcont ())
(define-record proc-valcont (rands env k))
(define-record all-argcont (proc k))
(define-record test-valcont (then-exp else-exp env k))
(define-record first-valcont (rands env k))
(define-record rest-argcont (first k))

(define apply-continuation
  (lambda (k val)
    (variant-case k     
      (final-valcont ()
        (let ((final val))
          final))
      (proc-valcont (rands env k)
        (let ((proc val))
          (eval-rands rands env 
            (make-all-argcont proc k))))
      (all-argcont (proc k) 
        (let ((all val))
          (apply-proc proc all k)))
      (test-valcont (then-exp else-exp env k)
        (let ((test val))
          (if (true-value? test)
              (eval-exp then-exp env k)
              (eval-exp else-exp env k))))
      (first-valcont (rands env k)
        (let ((first val))
          (eval-rands (cdr rands) env
            (make-rest-argcont first k))))
      (rest-argcont (first k)
        (let ((rest val))
          (apply-continuation k (cons first rest)))))))

;;; Anonymous figure : page 306

(define final-valcont (make-final-valcont))

;;; Anonymous figure : page 307, 308

(define apply-proc
  (lambda (proc args k)
    (variant-case proc
      (prim-proc (prim-op)
        (apply-continuation k (apply-prim-op prim-op args)))
      (closure (formals body env)
        (eval-exp body (extend-env formals args env) k))
      (abort () (apply-continuation final-valcont (car args)))
      (continuation (cont)
        (apply-continuation cont (car args)))
      (else (error "Invalid procedure:" proc)))))

