;;; $Id: lambda-multiple+if-examples.scm,v 1.4 1999/01/22 17:36:04 leavens Exp $

;;; This code works with the following grammar
;;;	<lcm+if-exp> ::= <varref>
;;;               | ( quote <symbol> )
;;;		  | ( lambda ( {<var>}* ) <lcm+if-exp> )
;;;		  | ( if <lcm+if-exp> <lcm+if-exp> <lcm+if-exp> )
;;;		  | ( <lcm+if-exp> {<lcm+if-exp>}* )
;;;     <var> ::= <symbol>
;;;     <varref> ::= <symbol>

(load-from-lib "lambda-multiple+if.scm")

(define count-varrefs
  ;; TYPE: (-> (lcm+if-exp) number)
  (lambda (exp)
    ;; ENSURES: result is the number of <varref>s in exp
    (cond
     ((varref? exp) 1)
     ((quote? exp) 0)
     ((lambda? exp) (count-varrefs (lambda->body exp)))
     ((if? exp) (+ (count-varrefs (if->test-exp exp))
		   (count-varrefs (if->then-exp exp))
		   (count-varrefs (if->else-exp exp))))
     ((app? exp) (+ (count-varrefs (app->rator exp))
		    (sum (map count-varrefs (app->rands exp)))))
     (else (error "count-varrefs, should not happen, syntax error: " exp)))))

(define sum
  ;; TYPE: (-> ((list number)) number)
  (lambda (lon)
    (if (null? lon)
	0
	(+ (car lon) (sum (cdr lon))))))

(define all-varrefs
  ;; TYPE: (-> (lcm+if-exp) (list symbol))
  (lambda (exp)
    ;; ENSURES: result is a list of all the <varref>s in exp
    ;;          (this list may contain duplicate occurrences of symbols).
    (cond
     ((varref? exp) (list (varref->var exp)))
     ((quote? exp) '())
     ((lambda? exp) (all-varrefs (lambda->body exp)))
     ((if? exp) (append (all-varrefs (if->test-exp exp))
			(all-varrefs (if->then-exp exp))
			(all-varrefs (if->else-exp exp))))
     ((app? exp) (append
		  (all-varrefs (app->rator exp))
		  (append-all (map all-varrefs (app->rands exp)))))
     (else (error "all-varrefs, should not happen, syntax error: " exp)))))

;;; Note: in class we did all-varrefs so it returned a set,
;;; so we used set-union instead of append.

(define append-all
  ;; TYPE: (-> ((list (list symbol))) (list symbol))
  (lambda (lls)
    (if (null? lls)
	'()
	(append (car lls) (append-all (cdr lls))))))

(define subst-varrefs
  ;; TYPE: (-> (symbol symbol lambda-1+quote-exp) lambda-1+quote-exp)
  (lambda (new old exp)
    ;; ENSURES: result is the same expression as exp, but with all varrefs
    ;;          that are the same as old replaced by new.
    (cond
     ((varref? exp)
      (if (eq? old (varref->var exp))
	  (make-varref new)
	  exp));; equal & slower: (make-varref (varref->var exp))
     ((quote? exp) exp);; equal & slower: (make-quote (quote->symbol exp))
     ((lambda? exp)
      (make-lambda
       (lambda->formals exp)
       (subst-varrefs new old (lambda->body exp))))
     ((if? exp)
      (make-if (subst-varrefs new old (if->test-exp exp))
	       (subst-varrefs new old (if->then-exp exp))
	       (subst-varrefs new old (if->else-exp exp))))
     ((app? exp)
      (make-app
       (subst-varrefs new old (app->rator exp))
       (map (lambda (e) (subst-varrefs new old e)) (app->rands exp))))
     (else (error "subst-varrefs, should not happen, syntax error: " exp)))))
