;;; $Id: list-structure-parser.scm,v 1.2 1999/08/11 16:28:09 leavens Exp $

;;; ****************************************************************

;;; Beginning of figures for Appendix B

;;; Figure B.1 : page 465

(define-record define (var exp))
(define-record varref (var))
(define-record lit (datum))
(define-record app (rator rands))
(define-record if (test-exp then-exp else-exp))
(define-record let (decls body))
(define-record decl (var exp))
(define-record proc (formals body))
(define-record varassign (var exp))
(define-record letmutable (decls body))
(define-record begin (exp1 exp2))
(define-record letrecproc (procdecls body))
(define-record procdecl (var formals body))
(define-record letrec (decls body))
(define-record dynassign (var exp body))
(define-record letdynamic (decls body))

;;; Figure B.2 : page 466

(define-record definearray (var dim-exp)) 
(define-record letarray (arraydecls body))
(define-record arrayref (array index))
(define-record arrayassign (array index exp))
(define-record letproc (procdecls body))
(define-record local (decls body))
(define-record keydecl (var exp))

;;; Figure B.2 : page 466 %%%%% This is for 3rd printing %%%%%%

(define-record definearray (var lex-exp))
(define-record letarray (arraydecls body))
(define-record arrayref (array index))
(define-record arrayassign (array index exp))
(define-record letproc (procdecls body))
(define-record local (decls body))
(define-record keydecl (var exp))

;;; Figure B.3 : page 466

(define-record super-meth-app (name rands))
(define-record meth-app (name rands))
(define-record i-varref (var))
(define-record c-varref (var))
(define-record i-varassign (var exp))
(define-record c-varassign (var exp))
(define-record method (formals body))
(define-record new-simpleinst (class-exp)) ; corrected for second printing
(define-record new-simpleclass (c-vars i-vars methdecls init-exp))
(define-record new-class (parent-exp c-vars i-vars methdecls init-exp))
(define-record new-instance (class-exp parent-exp i-vars methdecls))

;;; Figure B.4 : page 466

(define-record abort (exp))
(define-record letcont (var body))
(define-record callcc (exp))
(define-record coroutine (exp))
(define-record wind (pre body post))

;;; Figure B.5 : page 466

(define-record sum (rands))

;;; End of figures for Appendix B

;;; ****************************************************************

;;; Beginning of figures for Appendix H

;;; Figure H.1 : page 502, 503

(define parse
  (lambda (exp)
    (if (form? exp)
        (parse-form exp)
        (parse-exp exp))))

(define form?
  (lambda (x)
    (and (pair? x) (memq (car x) '(define definearray)))))

(define parse-form
  (lambda (form)
    (let ((form-pair (assq (car form) form-table)))
      (if (and (pair? form-pair) (approve form))
          (apply (cdr form-pair) (cdr form))
          (error "Invalid top-level form:" form)))))

(define parse-exp
  (lambda (exp)
    (cond
      ((symbol? exp) (make-varref exp))
      ((number? exp) (make-lit exp))
      ((and (pair? exp) (not (form? exp))) (parse-pair exp))
      (else (error "Invalid expression:" exp)))))

(define parse-pair
  (lambda (exp)
    (let ((x (car exp)))
      (if (symbol? x)
          (let ((form-pair (assq x form-table)))
            (if (and (pair? form-pair) (approve exp))
                (apply (cdr form-pair) (cdr exp))
                (make-app (parse-exp (car exp)) (parse-rands (cdr exp)))))
          (make-app (parse-exp (car exp)) (parse-rands (cdr exp)))))))

(define approve
  (lambda (x) #t))

(define parse-decl
  (lambda (decl)
    (make-decl (car decl) (parse-exp (cadr decl)))))

(define parse-procdecl
  (lambda (triple)
    (make-procdecl (car triple) (cadr triple) (parse-exp (caddr triple)))))

(define parse-with-keys
  (lambda (proc)
    (letrec ((loop (lambda (exps)
                     (cond
                       ((null? exps) '())
                       ((and (pair? (car exps)) (eq? (car (car exps)) ':))
                        (map (lambda (decl)
                               (make-keydecl (cadr decl)
                                 (parse-exp (caddr decl))))
                             exps))
                       (else (cons (proc (car exps)) (loop (cdr exps))))))))
      loop)))

(define parse-rands (parse-with-keys parse))
(define parse-formals (parse-with-keys (lambda (x) x)))

;;; Figure H.2 : page 504, 505, 506

(define form-table
  (list
    (cons 'define
      (lambda (var exp)
        (make-define var (parse-exp exp))))
    (cons 'definearray
      (lambda (var exp)
        (make-definearray var (parse-exp exp))))
    (cons 'if
      (lambda (test-exp then-exp else-exp)
        (make-if (parse-exp test-exp) (parse-exp then-exp) (parse-exp else-exp))))
    (cons 'let
      (lambda (decls body)
        (make-let (map parse-decl decls) (parse-exp body))))
    (cons 'proc
      (lambda (formals body)
        (make-proc (parse-formals formals) (parse-exp body))))
    (cons 'varassign
      (lambda (var body)
        (make-varassign var (parse-exp body))))
    (cons 'letmutable
      (lambda (decls body)
        (make-letmutable (map parse-decl decls) (parse-exp body))))
    (cons 'begin
      (letrec ((loop (lambda (exps)
                       (if (null? (cdr exps))
                           (parse-exp (car exps))
                           (make-begin (parse-exp (car exps)) (loop (cdr exps)))))))
        (lambda exps
          (loop exps))))
    (cons 'letrecproc
      (lambda (procdecls body)
        (make-letrecproc (map parse-procdecl procdecls) (parse-exp body))))
    (cons 'letrec
      (lambda (decls body) (make-letrec (map parse-decl decls) (parse-exp body))))
    (cons 'dynassign
      (lambda (var exp body)
        (make-dynassign var (parse-exp exp) (parse-exp body))))
    (cons 'letdynamic
      (lambda (decls body)
        (make-letdynamic (map parse-decl decls) (parse-exp body))))))

(define form-table-for-chapter-6
  (list 
    (cons 'letarray
      (lambda (arraydecls body)
        (make-letarray (map parse-decl arraydecls) (parse-exp body))))
    (cons 'arrayref
      (lambda (array index)
        (make-arrayref (parse-exp array) (parse-exp index))))
    (cons 'arrayassign
      (lambda (array index exp)
        (make-arrayassign (parse-exp array) (parse-exp index) (parse-exp exp))))
    (cons 'letproc
      (lambda (procdecls body)
        (make-letproc (map parse-procdecl procdecls)
          (parse-exp body))))
    (cons 'local
      (lambda (decls body)
        (make-local (map parse-decl decls) (parse-exp body))))))

(define form-table-for-chapter-7
  (list
    (cons 'i-varref
      (lambda (var) (make-i-varref var)))
    (cons 'c-varref
      (lambda (var) (make-c-varref var)))
    (cons 'i-varassign
      (lambda (var exp) (make-i-varassign var (parse-exp exp))))
    (cons 'c-varassign
      (lambda (var exp) (make-c-varassign var (parse-exp exp))))
    (cons 'method
      (lambda (formals body)
        (make-method (parse-formals formals) (parse-exp body))))
    (cons 'meth-app
      (lambda args
        (let ((name (car args)) (rands (cdr args)))
          (if (and (pair? rands) (symbol? (car rands)) (eq? (car rands) 'super))
              (make-super-meth-app name (parse-rands (cdr rands)))
              (make-meth-app name (parse-rands rands))))))
    (cons 'simpleinstance
      (lambda (class-exp)
        (make-new-simpleinst (parse-exp class-exp))))
    (cons 'simpleclass
      (lambda (c-vars i-vars methdecls init-exp)
        (make-new-simpleclass c-vars i-vars
          (map parse-decl methdecls)
          (parse-exp init-exp))))
    (cons 'class
      (lambda (parent-exp c-vars i-vars methdecls init-exp)
        (make-new-class (parse-exp parent-exp) c-vars i-vars
          (map parse-decl methdecls)
          (parse-exp init-exp))))
    (cons 'instance
      (lambda (class-exp parent-exp i-vars methdecls)
        (make-new-instance (parse-exp class-exp) (parse-exp parent-exp)
          i-vars
          (map parse-decl methdecls))))))

(define form-table-for-chapter-9
  (list     
    (cons 'abort
      (lambda (exp) (make-abort (parse-exp exp))))
    (cons 'letcont
      (lambda (var body) (make-letcont var (parse-exp body))))
    (cons 'callcc
      (lambda (exp) (make-callcc (parse-exp exp))))
    (cons 'coroutine
      (lambda (exp) (make-coroutine (parse-exp exp))))
    (cons 'wind
      (lambda (pre body post)
        (make-wind (parse-exp pre) (parse-exp body) (parse-exp post))))))

(define form-table-for-chapter-12
  (list
    (cons 'sum
      (lambda rands
        (make-sum (parse-rands rands))))))

;;; End of figures for Appendix H
