;;; $Id: num-of.scm,v 1.3 1999/01/22 07:59:38 leavens Exp $

;; Examples:
;; (num-of 'a (parse-sym-exp 'a)) ==> 1
;; (num-of 'a (parse-sym-exp 'b)) ==> 0
;; (num-of 'x (parse-sym-exp '(x y (x)))) ==> 2
;; (num-of 'x (parse-sym-exp '(y (x)))) ==> 1

;; Grammar:
;; <sym-exp>::= <symbol> | <s-list>
;; <s-list> ::= ( {<sym-exp>}* )

(load-from-lib "sym-exp.scm")

(define num-of
  ; TYPE: (-> (symbol sym-exp) number)
  (lambda (sym symexp)
    (if (sym-exp-symbol? symexp)
	(if (eq? sym (sym-exp->symbol symexp))
	    1
	    0)
	(num-of-slist sym (sym-exp->s-list symexp)))))


;; Examples:
;; (num-of-slist 'x (parse-s-list '())) ==> 0
;; (num-of-slist 'x (parse-s-list '(x y (x)))) ==> 2
;; (num-of-slist 'x (parse-s-list '(y (x)))) ==> 1

(define num-of-slist
  ; TYPE: (-> (symbol (list sym-exp)) number)
  (lambda (sym slist)
    (if (null? slist)
	0
	(+ (num-of sym (car slist))
	   (num-of-slist sym (cdr slist))))))
