;;; $Id: type-check-global-var-external-types.scm,v 1.1 1999/06/23 16:57:22 leavens Exp $
;;;
;;; External types for built-in Scheme procedures
;;; AUTHORS: Curtis Clifton and Gary T. Leavens,
;;;          from an earlier verison by Steven L. Jenkins.

(define *tc:pre-global-var-external-type-env*
  ;; TYPE: (env symbol external-type-expr)
  ;;
  ;; This is the table for external-type-exprs of built in expressions.
  ;; The order follows the listing in the body of R4RS, to make it possible
  ;; to check this list against that report.
  ;; Some additions from SCM and Com S 342 are found at the end.
  ;; Types marked "restrictive" are more restrictive than R4RS.
  ;; Types marked "unsound" make the type system unsound.
  (tc:assoc-list->env
   '(
     ;; 4.1.6
     (set!     . (-> (T T) void))	; unsound
     ;; 4.2.1
     (and      . (-> (boolean ...) boolean)) ; restrictive
     (or       . (-> (boolean ...) boolean)) ; restrictive
     ;; 4.2.5
     (delay    . (-> (T) (promise T)))
     ;; 6.1
     (not      . (-> (boolean) boolean))	; restrictive
     (boolean? . (-> (T) boolean))
     ;; 6.2
     (eqv?     . (and (-> (boolean boolean) boolean)
		      (-> (number number) boolean)
		      (-> (symbol symbol) boolean)
		      (-> (char char) boolean))) ; restrictive
     (eq?      . (and (-> (T T) boolean)
		      (-> (U symbol) boolean)
		      (-> (symbol V) boolean))) ; restrictive
     (equal?   . (and (-> (T T) boolean)
		      (-> (U symbol) boolean)
		      (-> (symbol V) boolean))) ; restrictive
     ;; 6.3
     (pair?    . (-> (T) boolean))
     (cons     . (and (-> (T (list T)) (list T))
		      (-> (U (list V)) (list datum))
		      (-> (S W) (pair S W))))
     (car      . (and (-> ((list T)) T)
		      (-> ((pair U V)) U)))
     (cdr      . (and (-> ((list T)) (list T))
		      (-> ((pair U V)) V)))
     (set-car! . (and (-> ((list T) T) void)
		      (-> ((pair U V) U) void)))
     (set-cdr! . (and (-> ((list T) (list T)) void)
		      (-> ((pair U V) V) void)))
     ;; caar, etc. are handled separartely
     (null?    . (-> ((list T)) boolean))       ; restrictive
     (list?    . (-> (T) boolean))
     (list     . (and (-> (T ...) (list T))
		      (-> (datum ...) (list datum))))
     (length   . (-> ((list T)) number))
     (append   . (and (-> ((list T) ...) (list T))
		      (-> ((list W) (list X) ...) (list datum))))
     (reverse  . (-> ((list T)) (list T)))
     (list-tail . (-> ((list T) number) (list T)))
     (list-ref . (-> ((list T) number) T))
     (memq     . (-> (T (list T)) boolean))  ; restrictive
     (memv     . (-> (T (list T)) boolean))  ; restrictive
     (member   . (-> (T (list T)) boolean))  ; restrictive
     (assq     . (and (-> (T (list (list T))) (list T))
		      (-> (V (list (pair V S))) (pair V S)))) ; restrictive
     (assv     . (and (-> (T (list (list T))) (list T))
		      (-> (V (list (pair V S))) (pair V S)))) ; restrictive
     (assoc    . (and (-> (T (list (list T))) (list T))
		      (-> (V (list (pair V S))) (pair V S)))) ; restrictive
     ;; 6.4
     (symbol?  . (-> (T) boolean))
     (symbol->string . (-> (symbol) string))
     (string->symbol . (-> (string) symbol))
     ;; 6.5.5
     (number?  . (-> (T) boolean))
     (complex? . (-> (T) boolean))
     (real?    . (-> (T) boolean))
     (rational? . (-> (T) boolean))
     (integer? . (-> (T) boolean))
     (exact?   . (-> (T) boolean))
     (inexact? . (-> (T) boolean))
     (=        . (-> (number ...) boolean))
     (<        . (-> (number ...) boolean))
     (>        . (-> (number ...) boolean))
     (<=       . (-> (number ...) boolean))
     (>=       . (-> (number ...) boolean))
     (zero?    . (-> (number) boolean))
     (negative? . (-> (number) boolean))
     (positive? . (-> (number) boolean))
     (odd?      . (-> (number) boolean))
     (even?     . (-> (number) boolean))
     (max       . (-> (number number  ...) number))     
     (min       . (-> (number number ...) number))
     (+         . (-> (number ...) number))
     (*         . (-> (number ...) number))
     (-         . (-> (number number ...) number))
     (/         . (-> (number number ...) number))
     (abs       . (-> (number) number))
     (quotient  . (-> (number number) number))     
     (remainder . (-> (number number) number))     
     (modulo    . (-> (number number) number))
     (gcd       . (-> (number ...) number))
     (lcm       . (-> (number ...) number))
     (numerator . (-> (number) number))
     (denominator . (-> (number) number))
     (rationalize . (-> (number number) number))
     (floor      . (-> (number) number))
     (ceiling    . (-> (number) number))
     (truncate   . (-> (number) number))     
     (round      . (-> (number) number))
     (exp        . (-> (number) number))
     (log        . (-> (number) number))          
     (sin        . (-> (number) number))     
     (cos        . (-> (number) number))     
     (asin       . (-> (number) number))     
     (acos       . (-> (number) number))
     (tan        . (-> (number) number))     
     (atan       . (and (-> (number) number)
			(-> (number number) number)))
     (expt       . (-> (number number) number))
     (sqrt       . (-> (number) number))
     (make-rectangular . (-> (number number) number))
     (make-polar . (-> (number number) number))
     (real-part  . (-> (number) number))
     (imag-part  . (-> (number) number))
     (magnitude  . (-> (number) number))
     (angle      . (-> (number) number))
     (exact->inexact . (-> (number) number))
     (inexact->exact . (-> (number) number))
     ;; 6.5.6
     (number->string . (and (-> (number) string)
			    (-> (number number) string)))
     (string->number . (and (-> (string) number)
			    (-> (string number) number)))
     ;; 6.6
     (char?       . (-> (T) boolean))
     (char=?      . (-> (char char) boolean))
     (char<?      . (-> (char char) boolean))
     (char>?      . (-> (char char) boolean))
     (char<=?     . (-> (char char) boolean))
     (char>=?     . (-> (char char) boolean))
     (char-ci=?   . (-> (char char) boolean))
     (char-ci<?   . (-> (char char) boolean))
     (char-ci>?   . (-> (char char) boolean))
     (char-ci<=?  . (-> (char char) boolean))
     (char-ci>=?  . (-> (char char) boolean))
     (char-alphabetic? . (-> (char) boolean))
     (char-numeric?    . (-> (char) boolean))
     (char-whitespace? . (-> (char) boolean))
     (char-upper-case? . (-> (char) boolean))
     (char-lower-case? . (-> (char) boolean))
     (char->integer    . (-> (char) number))
     (integer->char    . (-> (number) char))
     (char-upcase      . (-> (char) char))
     (char-downcase    . (-> (char) char))
     ;; 6.7
     (string?          . (-> (T) boolean))
     (make-string      . (and (-> (number char) string)
			      (-> (number) string))) ; unsound
     (string           . (-> (char ...) string))
     (string-length    . (-> (string) number))
     (string-ref       . (-> (string number) char))
     (string-set!      . (-> (string number char) void))     
     (string=?         . (-> (string ...) boolean))
     (string-ci=?      . (-> (string ...) boolean))
     (string<?         . (-> (string string) boolean))
     (string>?         . (-> (string string) boolean))
     (string<=?        . (-> (string string) boolean))
     (string>=?        . (-> (string string) boolean))
     (string-ci<?      . (-> (string string) boolean))
     (string-ci>?      . (-> (string string) boolean))
     (string-ci<=?     . (-> (string string) boolean))
     (string-ci>=?     . (-> (string string) boolean))
     (substring        . (-> (string number number) string))     
     (string-append    . (-> (string ...) string))
     (string->list     . (-> (string) (list char)))
     (list->string     . (-> ((list char)) string))
     (string-copy      . (-> (string) string))
     (string-fill!     . (-> (string) char))
     ;; 6.8
     (vector?          . (-> (T) boolean))
     (make-vector      . (and (-> (number T) (vector T))
			      (-> (number) (vector U)))) ; unsound
     (vector           . (and (-> (T ...) (vector T))
			      (-> (datum ...) (vector datum))))
     (vector-length    . (-> ((vector T)) number))
     (vector-ref       . (-> ((vector T) number) T))
     (vector-set!      . (-> ((vector T) number T) void))
     (vector->list     . (-> ((vector T)) (list T)))
     (list->vector     . (-> ((list T)) (vector T)))
     (vector-fill!     . (-> ((vector T) T) void))
     ;; 6.9
     (procedure?       . (-> (T) boolean))
     (apply            . (and (-> ((-> (D ...) E) (list D)) E)
			      (-> ((-> (T) U) (list T)) U)
			      (-> ((-> (F F) G) (list F)) G)
			      (-> ((-> (H H H) I) (list H)) I)
			      (-> ((-> (V S X ...) W)
				   (pair V (pair S (list X))))
				  W)
			      (-> ((-> (Y Z A C ...) B)
				   (pair Y (pair Z (pair A (list C)))))
				  B)
			      (-> ((-> (J ...) K) J ...) K)
			      ))		; restrictive
     (map      . (and (-> ((-> (T) U) (list T)) (list U))
		      (-> ((-> (W X) Y)
			   (list W) (list X))
			  (list Y))
		      (-> ((-> (Z A B) C)
			   (list Z) (list A) (list B))
			  (list C))
		      (-> ((-> (D D ...) E) (list D) (list D) ...)
			  (list E))
		      ))			; restrictive
     (for-each . (and (-> ((-> (T) void) (list T)) void)
		      (-> ((-> (W X) void)
			   (list W) (list X))
			  void)
		      (-> ((-> (Z A B) void)
			   (list Z) (list A) (list B))
			  void)
		      (-> ((-> (D ...) void) (list D) ...) void)
		      ))			; restrictive
     (force     . (-> ((promise T)) T))
     (call-with-current-continuation  . (-> ((-> ((-> (T) poof)) V)) T))
     (call/cc   . (-> ((-> ((-> (T) poof)) V)) T))
     ;; 6.10.1
     (call-with-input-file  . (-> (string (-> (port) T)) T))
     (call-with-output-file . (-> (string (-> (port) T)) T))
     (input-port?           . (-> (T) boolean))
     (output-port?          . (-> (T) boolean))
     (current-input-port    . (-> () port))
     (current-output-port   . (-> () port))
     (with-input-from-file  . (-> (string (-> () T)) T))
     (with-output-to-file   . (-> (string (-> () T)) T))
     (open-input-file       . (-> (string) port))
     (open-output-file      . (-> (string) port))
     (close-input-port      . (-> (port) void))
     (close-output-port     . (-> (port) void))
     ;; 6.10.2
     (read       . (and (-> () T)	; unsound
			(-> (port) T)))	; unsound
     (read-char  . (and (-> () char)
			(-> (port) char)))
     (peek-char  . (and (-> () char)
			(-> (port) char))) ; restrictive
     (eof-object? . (-> (T) boolean))
     (char-ready? . (-> (port) boolean))
     ;; 6.10.3
     (write       . (and (-> (T) void)
			 (-> (U port) void)))
     (display     . (and (-> (T) void)
			 (-> (U port) void)))
     (newline     . (and (-> () void)
			 (-> (port) void)))
     (write-char  . (and (-> (char) void)
			 (-> (char port) void)))
     ;; 6.10.4
     (load           . (-> (string) void))
     (transcript-on  . (-> (string) void))
     (transcript-off . (-> () void))
     ;; additions in SCM
     (eval           . (-> (datum) T))	; unsound
     (error          . (and (-> () poof)
			    (-> (T) poof)
			    (-> (U V) poof)
			    (-> (S W X) poof)
			    (-> (Y Z A B datum ...) poof)))
     (exit           . (-> () poof))
     (file-exists?   . (-> (string) boolean))
     (force-output   . (-> () void))
     (require        . (-> (symbol) void))
     (verbose        . (-> (number) void))
     (quit           . (-> () poof))
     ;; additions for 342
     (writeln        . (and (-> () void)    ; from (load-from-lib "writeln")
			    (-> (T) void)
			    (-> (U V) void)
			    (-> (S W X) void)
			    (-> (Y Z A B datum ...) void)))
     
     (displayln      . (and (-> () void)      ; from (load-from-lib "displayln")
			    (-> (T) void)
			    (-> (U V) void)
			    (-> (S W X) void)
			    (-> (Y Z A B datum ...) void)))
     (load-from-lib  . (-> (string) void))     ; from (load "load-from-lib")
     (load-quietly-from-lib  . (-> (string) void))
     (load-quietly   . (-> (string) void))
     (random         . (-> (number) number))  ; from (require 'random)
     (trace          . (-> (datum ...) void)) ; from (require 'trace)
     (untrace        . (-> (datum ...) void))
     (pretty-print   . (-> (datum) void))   ; from (require 'pretty-print)
     (*eval-even-if-error* . boolean)     ; from the type checker
     (type-prompting-on! . (-> () void))	; from the type checker
     (type-prompting-off! . (-> () void)) ; from the type checker    
     (type-prompting-on? . (-> () boolean)) ; from the type checker
     (type-tracing-set! . (-> (number) void)) ; from the type checker
     (type-tracing-off! . (-> () void))	; from the type checker    
     (type-tracing-level . (-> () number)) ; from the type checker
     (type-help . (-> () void))		; from the type checker
     (test-hw1       . (-> (string) void))  ; from (load-from-lib "test-homework")
     (test-hw2       . (-> (string) void))
     (test-hw3       . (-> (string) void))
     (test-hw4       . (-> (string) void))
     (test-hw5       . (-> (string) void))
     (test-hw6       . (-> (string) void))
     (test-hw7       . (-> (string) void))
     (test-hw8       . (-> (string) void))
     (test-hw9       . (-> (string) void))
     (test-hw10       . (-> (string) void))
     (test-hw11       . (-> (string) void))
     (test-hw12       . (-> (string) void))
     (run-test-case-maker . (-> ((-> (datum) void) (-> (datum) void))
				(-> (datum) void)))
     (run-test-case   . (-> (datum) void))
     (regression-test-maker . (-> ((-> (S T) boolean)
				   (-> (T) void)
				   (-> (T) void))
				  (-> (T symbol T) void)))
     (run-regression-test . (-> (scheme-exp symbol datum) void))
     (run-regression-tests-maker . (-> ((-> (datum datum) boolean)
					(-> (datum) void)
					(-> (datum) void))
				       (-> ((list datum)) void)))
     (run-regression-tests-equal? . (-> ((list datum)) void))
     )))
