;;; $Id: type-check-variable.scm,v 1.8 1999/06/23 16:57:22 leavens Exp $
;;; Type inference for Scheme programs
;;; AUTHORS: Curtis Clifton and Gary T. Leavens, 
;;;          from an earlier verison by Steven L. Jenkins.

;;; Logical variables (universally quantified) used in type checking.

;;; Logical variables are represented by symbols starting with a "?" (e.g. ?a)

(define tc:variable? 
  ;; TYPE: (-> (datum) boolean)
  (lambda (x)
    (and (symbol? x)
	 (char=? (string-ref (symbol->string x) 0) #\?))))

(define tc:symbol->variable
  ;; TYPE: (-> (symbol) variable)
  (lambda (s)
    (if (tc:variable? s)
	s
	(error "tc:symbol->variable passed bad symbol" s))))

(define tc:variable->symbol
  ;; TYPE: (-> (symbol) variable)
  (lambda (s) s))

(define tc:new-variable 
  ;; TYPE: (-> () variable)
  (let ((variable-count 0))
    (lambda ()
      (set! variable-count (+ variable-count 1))
      (string->symbol (string-append "?" (number->string variable-count))))))

(define tc:variable-new-level
  ;; TYPE: (-> (number) (-> (variable) variable))
  (lambda (level)
    (lambda (var)
      (string->symbol
       (string-append
	(tc:before-suffix (symbol->string var))
	"#"
	(number->string level))))))

(define tc:delimiter-pos
  ;; TYPE: (-> (string) number)
  (lambda (s)
    (letrec ((findit ; TYPE: (-> (number) number)
	      (lambda (n)
		(if (< n 0)
		    -1
		    (if (equal? (string-ref s n) #\#)
			n
			(findit (- n 1)))))))
      (findit (- (string-length s) 1)))))

(define tc:before-suffix
  ;; TYPE: (-> (string) string)
  (lambda (s)
    ;; ENSURES: result is the prefix of s
    ;;          that occurs before the tc:first '#' character in s
    (let ((i (tc:delimiter-pos s)))
      (if (< i 0)
	  s
	  (substring s 0 i)))))

